"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.OpenApiSpecificationBuilder = void 0;
const path_builder_1 = require("./path-builder");
class OpenApiSpecificationBuilder {
    oas;
    parts = {
        schemes: [],
        apiIntegrations: []
    };
    constructor(oas) {
        this.oas = oas;
    }
    _defaultResponses = {};
    build() {
        return this.oas;
    }
    withAWSCognitoSecurityScheme(key, userPoolEndpoint, clientId, identitySource = '$request.header.Authorization') {
        const scheme = {
            type: "oauth2",
            "x-amazon-apigateway-authorizer": {
                type: "jwt",
                jwtConfiguration: {
                    issuer: userPoolEndpoint,
                    audience: [clientId]
                },
                identitySource: identitySource
            }
        };
        this.parts.schemes.push(key);
        return this.addComponent('securitySchemes', () => ({ [key]: scheme }));
    }
    withAWSLambdaApiGatewayIntegration(key, functionUri, payloadFormatVersion = '1.0', passthroughBehavior = 'when_no_templates') {
        const scheme = {
            type: "aws_proxy",
            httpMethod: "POST",
            uri: functionUri,
            passthroughBehavior: passthroughBehavior,
            payloadFormatVersion: payloadFormatVersion
        };
        this.parts.apiIntegrations.push(key);
        return this.addComponent('x-amazon-apigateway-integrations', () => ({ [key]: scheme }));
    }
    get securitySchemes() {
        const schemes = this.parts.schemes.reduce((prev, key) => {
            return { ...prev, [key]: ((scopes) => {
                    return { [key]: scopes ?? [] };
                }) };
        }, {});
        return schemes;
    }
    get awsLambdaApiGatewayIntegration() {
        const apiIntegrations = this.parts.apiIntegrations.reduce((prev, key) => {
            return { ...prev, [key]: { "$ref": `#/components/x-amazon-apigateway-integrations/${key}` } };
        }, {});
        return apiIntegrations;
    }
    withPath(name, builder) {
        const pathBuilder = builder(new path_builder_1.PathBuilder('/' + name), this);
        this.oas.paths = { ...(this.oas.paths ?? {}), ...pathBuilder.path };
        return this;
    }
    jsonContent(key, example, examples, encoding) {
        return { 'application/json': this.media(key, example, examples, encoding) };
    }
    textContent(example, examples, encoding) {
        return { 'application/text': { schema: { type: 'string' }, example, examples, encoding } };
    }
    response(content, description = '', headers = []) {
        return { description, content, headers: headers.reduce((prev, name) => ({ ...prev, [name]: { required: true, schema: { type: 'string' } } }), {}) };
    }
    textResponse(description = '', headers = [], example, examples, encoding) {
        return this.response(this.textContent(example, examples, encoding), description, headers);
    }
    jsonResponse(key, description = '', headers = [], example, examples, encoding) {
        return this.response(this.jsonContent(key, example, examples, encoding), description, headers);
    }
    responseReference(key) {
        return this.componentReference('responses', key);
    }
    media(key, example, examples, encoding) {
        return { schema: this.reference(key), example, examples, encoding };
    }
    basicAuthScheme(description = '') {
        return {
            type: 'http',
            scheme: 'basic',
            description,
        };
    }
    apiKeyScheme(in_, name, description = '') {
        return {
            type: 'apiKey',
            in: in_,
            name,
            description,
        };
    }
    bearerScheme(description = '') {
        return {
            type: 'http',
            scheme: 'bearer',
            description
        };
    }
    oAuth2Scheme(flows, description = '') {
        return {
            type: 'oauth2',
            flows,
            description
        };
    }
    openIdConnectScheme(openIdConnectUrl, description = '') {
        return {
            type: 'openIdConnect',
            openIdConnectUrl,
            description
        };
    }
    query(name, required = true, multiple = false) {
        return this.parameter(name, "query", required, multiple ? { type: "array", items: { type: "string" } } : { type: "string" });
    }
    header(name, required = true) {
        return this.parameter(name, "header", required);
    }
    parameter(name, location, required = true, schema = { type: "string" }) {
        return {
            name,
            in: location,
            required,
            schema
        };
    }
    reference(key) {
        return this.componentReference('schemas', key);
    }
    componentReference(componentKey, key) {
        return { '$ref': `#/components/${componentKey}/${key}` };
    }
    defaultResponses(itemBuilder) {
        this._defaultResponses = itemBuilder(this);
        return this;
    }
    addComponent(location, itemBuilder) {
        const item = itemBuilder(this);
        const current = this.oas.components ?? {};
        if (typeof item === "object") {
            if (Array.isArray(item)) {
                this.oas = {
                    ...this.oas,
                    components: { ...current, [location]: [...(current[location] ?? []), ...item] }
                };
            }
            else {
                this.oas = {
                    ...this.oas,
                    components: { ...current, [location]: { ...(current[location] ?? {}), ...item } }
                };
            }
        }
        else {
            this.oas = { ...this.oas, components: { ...current, [location]: item } };
        }
        return this;
    }
    modifyOperationWithDefaultResponses(operation) {
        if (operation) {
            const { responses, ...rest } = operation;
            return { ...rest, responses: { ...this._defaultResponses, ...responses } };
        }
    }
    addDefaultResponsesToPath(path) {
        if (path.$ref)
            return path;
        const { get, put, post, delete: deleteOp, options, head, patch, trace, ...rest } = path;
        return {
            ...rest,
            ...(get ? { get: this.modifyOperationWithDefaultResponses(get) } : {}),
            ...(put ? { put: this.modifyOperationWithDefaultResponses(put) } : {}),
            ...(post ? { post: this.modifyOperationWithDefaultResponses(post) } : {}),
            ...(deleteOp ? { delete: this.modifyOperationWithDefaultResponses(deleteOp) } : {}),
            ...(options ? { options: this.modifyOperationWithDefaultResponses(options) } : {}),
            ...(head ? { head: this.modifyOperationWithDefaultResponses(head) } : {}),
            ...(patch ? { patch: this.modifyOperationWithDefaultResponses(patch) } : {}),
            ...(trace ? { trace: this.modifyOperationWithDefaultResponses(trace) } : {}),
            ...rest
        };
    }
    addDefaultResponses(paths) {
        return Object.keys(paths).reduce((previousValue, path) => ({ ...previousValue, [path]: this.addDefaultResponsesToPath(paths[path]) }), {});
    }
    add(location, itemBuilder) {
        const item = itemBuilder(this);
        if (typeof item === "object") {
            if (Array.isArray(item)) {
                this.oas = { ...this.oas, [location]: [...(this.oas[location] ?? []), ...item] };
            }
            else {
                this.oas = { ...this.oas, [location]: { ...(this.oas[location] ?? {}), ...(location === "paths" ? this.addDefaultResponses(item) : item) } };
            }
        }
        else {
            this.oas = { ...this.oas, [location]: item };
        }
        return this;
    }
    static create(schemas, info) {
        return new OpenApiSpecificationBuilder({ openapi: '3.0.0', info, paths: {}, ...schemas });
    }
}
exports.OpenApiSpecificationBuilder = OpenApiSpecificationBuilder;
