"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Validator = void 0;
const sdk_mapper_js_1 = require("./sdk-mapper.js");
class SchemaError {
    location;
    message;
    constructor(location, message) {
        this.location = location;
        this.message = message;
    }
}
/**
 * https://json-schema.org/draft/2019-09/json-schema-validation.html
 */
class Validator {
    static constValidation(value, constValue, location, schema) {
        const valid = constValue === undefined || value === constValue;
        return validate(value, location, schema, () => `Expected value to exactly equal ${constValue}`, valid);
    }
    static enumValidation(value, enumValues, location, schema) {
        const valid = enumValues === undefined || (Array.isArray(enumValues) && enumValues.includes(value));
        return validate(value, location, schema, () => `Expected value to exactly one of [${enumValues.join(', ')}]`, valid);
    }
    static validateGlobal(value, location, schema) {
        return [
            this.constValidation(value, schema.current.const, location, schema),
            this.enumValidation(value, schema.current.enum, location, schema)
        ].flat();
    }
    static actualSchema(schema, current) {
        if (!current['$ref'])
            return current;
        return (0, sdk_mapper_js_1.traversePath)(current['$ref'], schema);
    }
    static validateUnknown(value, location, schema) {
        const current = this.actualSchema(schema.schema, schema.current);
        const info = { schema: schema.schema, current };
        switch (current.type) {
            case 'string': return StringValidator.validate(value, location, info);
            case 'integer':
            case 'number': return NumberValidator.validate(value, location, info);
            case 'array': return ArrayValidator.validate(value, location, info);
            case 'boolean': return BooleanValidator.validate(value, location, info);
            case 'null': return NullValidator.validate(value, location, info);
            default: {
                if (current.anyOf) {
                    if (current.anyOf.find(anySchema => Validator.validate(value, anySchema, schema.schema).length === 0)) {
                        return [];
                    }
                    else {
                        return [{ value, schema, location, message: 'Failed to satisfy any of the subschemas' }];
                    }
                }
                return ObjectValidator.validate(value, location, info);
            }
        }
    }
    static validate(value, schema, parentSchema) {
        return this.validateUnknown(value, '#', { schema: parentSchema ?? schema, current: schema });
    }
}
exports.Validator = Validator;
class NumberValidator {
    static numberValidation(value, location, schema) {
        const valid = typeof value === 'number';
        return validate(value, location, schema, () => `Expected value to be a number`, valid);
    }
    static integerValidation(value, location, schema) {
        const valid = schema.current.type !== 'integer' || Number.isInteger(value);
        return validate(value, location, schema, () => `Expected value to be an integer`, valid);
    }
    static numericValidation(key, value, location, schema, validation, message) {
        if (schema.current[key] === undefined)
            return [];
        if (typeof schema.current[key] === 'number') {
            return validate(value, location, schema, () => message(schema.current[key]), validation(schema.current[key]));
        }
        throw new SchemaError(`${location}/${key}`, 'must be a number');
    }
    static multipleOfValidation(value, location, schema) {
        return this.numericValidation('multipleOf', value, location, schema, multipleOf => multipleOf > 0 && value % multipleOf === 0, multipleOf => `Expected value to be a multiple of ${multipleOf} and ${multipleOf} to be greater than 0`);
    }
    static maximumValidation(value, location, schema) {
        return this.numericValidation('maximum', value, location, schema, maximum => value <= maximum, maximum => `Expected value to be less than or equal to ${maximum}`);
    }
    static exclusiveMaximumValidation(value, location, schema) {
        return this.numericValidation('exclusiveMaximum', value, location, schema, exclusiveMaximum => value < exclusiveMaximum, exclusiveMaximum => `Expected value to be less than ${exclusiveMaximum}`);
    }
    static minimumValidation(value, location, schema) {
        return this.numericValidation('minimum', value, location, schema, exclusiveMinimum => value >= exclusiveMinimum, exclusiveMinimum => `Expected value to be greater than or equal to ${exclusiveMinimum}`);
    }
    static exclusiveMinimumValidation(value, location, schema) {
        return this.numericValidation('exclusiveMinimum', value, location, schema, exclusiveMinimum => value > exclusiveMinimum, exclusiveMinimum => `Expected value to be greater than ${exclusiveMinimum}`);
    }
    static validate(value, location, schema) {
        const isNumber = this.numberValidation(value, location, schema);
        if (isNumber.length === 0) {
            return [
                Validator.validateGlobal(value, location, schema),
                this.integerValidation(value, location, schema),
                this.multipleOfValidation(value, location, schema),
                this.maximumValidation(value, location, schema),
                this.exclusiveMaximumValidation(value, location, schema),
                this.minimumValidation(value, location, schema),
                this.exclusiveMinimumValidation(value, location, schema)
            ].flat();
        }
        return isNumber;
    }
}
class StringValidator {
    static stringValidation(value, location, schema) {
        const valid = typeof value === 'string';
        return validate(value, location, schema, () => `Expected value to be a string`, valid);
    }
    static maxLengthValidation(value, location, schema) {
        return NumberValidator.numericValidation('maxLength', value, location, schema, maxLength => maxLength > 0 && Number.isInteger(maxLength) && value.length <= maxLength, maxLength => `Expected value to have length less than or equal to ${maxLength}`);
    }
    static minLengthValidation(value, location, schema) {
        return NumberValidator.numericValidation('minLength', value, location, schema, minLength => minLength > 0 && Number.isInteger(minLength) && value.length >= minLength, minLength => `Expected value to have length greater than or equal to ${minLength}`);
    }
    static patternValidation(value, location, schema) {
        if (schema.current.pattern === undefined)
            return [];
        const pattern = new RegExp(schema.current.pattern);
        const isValid = value.match(pattern);
        return validate(value, location, schema, () => `Expected value to match pattern /${schema.current.pattern}/`, isValid);
    }
    static validate(value, location, schema) {
        const isString = this.stringValidation(value, location, schema);
        if (isString.length === 0) {
            return [
                Validator.validateGlobal(value, location, schema),
                this.maxLengthValidation(value, location, schema),
                this.minLengthValidation(value, location, schema),
                this.patternValidation(value, location, schema)
            ].flat();
        }
        return isString;
    }
}
class BooleanValidator {
    static booleanValidation(value, location, schema) {
        const valid = typeof value === 'boolean';
        return validate(value, location, schema, () => `Expected value to be a boolean`, valid);
    }
    static validate(value, location, schema) {
        const isBool = this.booleanValidation(value, location, schema);
        if (isBool.length === 0) {
            return Validator.validateGlobal(value, location, schema);
        }
        return isBool;
    }
}
class NullValidator {
    static nullValidation(value, location, schema) {
        const valid = value === null;
        return validate(value, location, schema, () => `Expected value to be null`, valid);
    }
    static validate(value, location, schema) {
        const isNull = this.nullValidation(value, location, schema);
        if (isNull.length === 0) {
            return Validator.validateGlobal(value, location, schema);
        }
        return isNull;
    }
}
class ArrayValidator {
    static arrayValidation(value, location, schema) {
        const valid = Array.isArray(value);
        return validate(value, location, schema, () => `Expected value to be an array`, valid);
    }
    static itemsValidation(value, location, schema) {
        if (schema.current.items === undefined)
            return [];
        if (Array.isArray(schema.current.items)) {
            const itemValidation = value.slice(0, schema.current.items.length).flatMap((item, index) => Validator.validateUnknown(item, `${location}/${index}`, { schema: schema.schema, current: schema.current.items[index] }));
            if (value.length > schema.current.items.length && schema.current.additionalItems) {
                return [...itemValidation, ...value.slice(schema.current.items.length).flatMap((item, index) => Validator.validateUnknown(item, `${location}/${schema.current.items.length + index}`, { schema: schema.schema, current: schema.current.additionalItems }))];
            }
            return itemValidation;
        }
        return value.flatMap((item, index) => Validator.validateUnknown(item, `${location}/${index}`, { schema: schema.schema, current: schema.current.items }));
    }
    static containsValidation(value, location, schema) {
        if (schema.current.contains === undefined)
            return [];
        if (value.find(it => Validator.validateUnknown(it, '', schema.current.contains).length === 0)) {
            return [];
        }
        return validate(value, location, schema, () => `Expected array to contain at least one item complying with ${schema.current.contains}`, false);
    }
    static uniqueValidation(value, location, schema) {
        if (!schema.current.uniqueItems)
            return [];
        if (new Set(value).size === value.length)
            return [];
        return validate(value, location, schema, () => `Expected array items to all be unique`, false);
    }
    static maxItemsValidation(value, location, schema) {
        return NumberValidator.numericValidation('maxItems', value, location, schema, maxItems => maxItems >= 0 && Number.isInteger(maxItems) && value.length <= maxItems, maxItems => `Expected array to have length less than or equal to ${maxItems}`);
    }
    static minItemsValidation(value, location, schema) {
        return NumberValidator.numericValidation('minItems', value, location, schema, minItems => minItems >= 0 && Number.isInteger(minItems) && value.length >= minItems, minItems => `Expected array ${value} to have length greater than or equal to ${minItems}`);
    }
    static validate(value, location, schema) {
        const isArray = this.arrayValidation(value, location, schema);
        if (isArray.length === 0) {
            return [
                Validator.validateGlobal(value, location, schema),
                this.itemsValidation(value, location, schema),
                this.containsValidation(value, location, schema),
                this.uniqueValidation(value, location, schema),
                this.maxItemsValidation(value, location, schema),
                this.minItemsValidation(value, location, schema),
            ].flat();
        }
        return isArray;
    }
}
class ObjectValidator {
    static objectValidation(value, location, schema) {
        const valid = value !== null && typeof value === 'object' && !Array.isArray(value);
        return validate(value, location, schema, () => `Expected value to be an object`, valid);
    }
    static minPropertiesValidation(value, location, schema) {
        const properties = Object.keys(value).length;
        return NumberValidator.numericValidation('minProperties', value, location, schema, minProperties => minProperties >= 0 && Number.isInteger(minProperties) && properties >= minProperties, minProperties => `Expected object ${value} to have min properties of ${minProperties} but had ${properties}`);
    }
    static maxPropertiesValidation(value, location, schema) {
        const properties = Object.keys(value).length;
        return NumberValidator.numericValidation('maxProperties', value, location, schema, maxProperties => maxProperties >= 0 && Number.isInteger(maxProperties) && properties <= maxProperties, maxProperties => `Expected object ${value} to have max properties of ${maxProperties} but had ${properties}`);
    }
    static requiredValidation(value, location, schema) {
        if (!schema.current.required)
            return [];
        const keys = Object.keys(value);
        const requiredMissing = schema.current.required.flatMap(key => {
            if (!keys.includes(key))
                return [key];
            return [];
        });
        if (requiredMissing.length === 0)
            return [];
        return [{ schema: schema.current, location, value, message: `Expected the following keys that were not present: [${requiredMissing.join(', ')}]` }];
    }
    static propertyValidation(value, location, schema) {
        if (!schema.current.properties || Object.keys(schema.current.properties).length === 0)
            return [];
        return Object.keys(schema.current.properties).flatMap(property => {
            const subElement = value[property];
            if (subElement === undefined)
                return [];
            return Validator.validateUnknown(subElement, `${location}/${property}`, { schema: schema.schema, current: schema.current.properties[property] });
        });
    }
    static additionalPropertyValidation(value, location, schema) {
        const aps = schema.current.additionalProperties;
        const ps = Object.keys(schema.current.properties ?? {});
        if (aps === undefined || aps === true)
            return [];
        const excessKeys = Object.keys(value ?? {}).filter(it => !ps.includes(it));
        if (excessKeys.length === 0)
            return [];
        if (aps === false)
            return [{ value, location, schema: schema.current, message: `Expected no additional properties but found the following: [${excessKeys.join(', ')}]` }];
        return excessKeys.flatMap(property => {
            const subElement = value[property];
            return Validator.validateUnknown(subElement, `${location}/${property}`, { schema: schema.schema, current: aps });
        });
    }
    static validate(value, location, schema) {
        const isObject = this.objectValidation(value, location, schema);
        if (isObject.length === 0) {
            return [
                Validator.validateGlobal(value, location, schema),
                this.minPropertiesValidation(value, location, schema),
                this.maxPropertiesValidation(value, location, schema),
                this.requiredValidation(value, location, schema),
                this.propertyValidation(value, location, schema),
                this.additionalPropertyValidation(value, location, schema),
            ].flat();
        }
        return isObject;
    }
}
function validate(value, location, schema, message, valid) {
    return valid ? [] : [{ location, message: message(), schema: schema.current, value }];
}
